/*
*******************************************************************************
*
* File:         main_ppu.c
* RCS:          $Header: $
* Description:  Main file of the PPU 
* Authors:      Fabrizio Petrini        
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/

#include "../common.h"
#include "runtime_ppu.h"


/*
*******************************************************************************
*
*                                  defines
*
*******************************************************************************
*/


#define MAX_FILE_NAME_LEN                    256


/*
*******************************************************************************
*
*                                 globals
*
*******************************************************************************
*/


/* number of SPUs */
unsigned N = 0;

/* when set, dynamically allocates a dma buffer */
int ALLOCATE_DYNAMIC_DMA_BUFFER = False;

/* the dma buffer size in bytes */
int DYNAMIC_DMA_BUFFER_SIZE = 0;

/* when set use huge pages; must be set in conjunction with
   ALLOCATE_DYNAMIC_DMA_BUFFER */
int USE_HUGE_TLB = False;

/* the file name to use huge pages */
char huge_filename[MAX_FILE_NAME_LEN] = "/huge/hugefile";

/* dynamically allocated communication buffer */
char* dynamic_dma_buffer = 0;

/* limit for the dynamically allocated communication buffer */
char* dynamic_dma_buffer_limit;

/* user control block */
control_block_t user_cb align128;


/*
*******************************************************************************
*
*                           huge page tables
*
*******************************************************************************
*/


void allocate_huge_buffer()
{
  int fmem;
  
  if ( !ALLOCATE_DYNAMIC_DMA_BUFFER )
    return;
  else if ( USE_HUGE_TLB )
    {
      /* Create a large contiguous memory buffer by allocating a large
	 page (or more); large page memory will also reduce the TLB
	 thrashing */
      if ( ( fmem = open( huge_filename, O_CREAT | O_RDWR, 0755) ) == -1 ) 
	{
	  printf( "Unable to open file %s (errno=%d). Using malloc heap.\n", 
		  huge_filename, errno );
	  dynamic_dma_buffer = 
	    (char *) memalign( 128, DYNAMIC_DMA_BUFFER_SIZE );
	} 
      else 
	{
	  /* delete the file so that huge pages will get freed on
	     program termination */
	  remove( huge_filename );
	  dynamic_dma_buffer = (char *) 
	    mmap( 0, DYNAMIC_DMA_BUFFER_SIZE, PROT_READ | PROT_WRITE, 
		  MAP_SHARED, fmem, 0);
	  
	  if ( dynamic_dma_buffer == MAP_FAILED ) 
	    {
	      printf( "Unable to mmap %s (errno=%d). Using malloc heap.\n", 
		      huge_filename, errno);
	      close( fmem );
	      dynamic_dma_buffer = 
		(char *) memalign( 128, DYNAMIC_DMA_BUFFER_SIZE );
	    }
	  else 
	    {
	      if ( (unsigned int) dynamic_dma_buffer & 0x7f )
		{
		  printf( "Dynamic buffer 0x%x is not 128-byte aligned\n",
			  (unsigned int) dynamic_dma_buffer ); 
		  
		  /* realign the buffer */
		  while ( (unsigned int) dynamic_dma_buffer & 0x7f ) 
		    dynamic_dma_buffer++;
		}
	      
	      printf( "Allocated huge tlb file addr 0x%x\n", 
		      (unsigned int) dynamic_dma_buffer );
	    }
	}
    }
  else
    /* standard malloc */
    dynamic_dma_buffer =(char *) memalign( 128, DYNAMIC_DMA_BUFFER_SIZE );
  
  if ( !dynamic_dma_buffer )
    {
      printf( "Unable to allocate dma buffer\n" ); 
      exit(0);
    }

  dynamic_dma_buffer_limit = dynamic_dma_buffer + DYNAMIC_DMA_BUFFER_SIZE;
  
  DB(printf( "0x%x 0x%x\n", (unsigned int) dynamic_dma_buffer,  
     (unsigned int) dynamic_dma_buffer + DYNAMIC_DMA_BUFFER_SIZE );)
}


/*
*******************************************************************************
*
*                        input parameters and usage  
*
*******************************************************************************
*/


void usage( const char * program_name /* i.e. argv[0] */)
{  
  printf ("Usage: %s [flags]\n", program_name);
  printf ("\n");
  printf (" Sizes can be expressed as k/K and m/M\n");  
  printf (" Flags may be any of\n");  
  printf ("  -H huge_filename         huge pages file name\n");
  printf ("  -n spus                  the number of spus\n");
  printf ("  -m array_size            size of the huge pages array\n"); 
  printf ("\n");
  exit (0);
}


int get_size( char *str )
{
  int  size;
  char mod[32];

  switch (sscanf (str, "%d%1[mMkK]", &size, mod))
  {
    case 1:
      return (size);
    case 2:
      switch (*mod)
      {
        case 'm':
        case 'M':
          return (size << 20);
        case 'k':
        case 'K':
          return (size << 10);
        default:
          return (size);
      }
    default:
      return (-1);
  }
}

/* parse the command line */
void parse_input( int argc, char** argv )
{
  char *prog_name;
  int c;

  /* extract the program name */
  for( prog_name = argv[0] + strlen (argv[0]);
      prog_name > argv[0] && *(prog_name - 1) != '/';
      prog_name-- )
    ;

  while( (c = getopt(argc, argv, "n:m:H:?")) 
	 != -1 )
    switch (c) 
      {
      case 'n':
        if ( ( N = get_size( optarg ) ) <= 0 ) 
	  {
	    fprintf( stderr, 
		     "Error: the number of spus must be a positive"
		     " number: %d\n", N );
	    usage(prog_name);
	  }

        if ( N > MAX_SPU ) 
	  {
	    fprintf( stderr, "Error: too many SPUs requested %d, max %d\n", 
		     N, MAX_SPU );
	    usage(prog_name);
	  }
        break;
      case 'm':
        if ( ( DYNAMIC_DMA_BUFFER_SIZE = get_size( optarg ) ) < 0 )
	  {
	    printf( "Dma buffer size must be a non-negative number: %d\n", 
		    DYNAMIC_DMA_BUFFER_SIZE );
	    usage( prog_name );
	  }
	
        if ( DYNAMIC_DMA_BUFFER_SIZE % ( 256 * 1024 ) )
	  {
	    printf( "Dma buffer size must be a multiple of 256k: %d\n\n", 
		    DYNAMIC_DMA_BUFFER_SIZE );
	    usage( prog_name );
	  }
	
        ALLOCATE_DYNAMIC_DMA_BUFFER = TRUE;
        break;
      case 'H':
        strcpy( huge_filename, optarg);
        USE_HUGE_TLB = TRUE;
        break;
      case 'h':
      case '?':
      default:
        usage(prog_name);
    }

  if ( !N ) 
    {
      fprintf( stderr, "Error: number of SPEs is 0\n");
      usage(prog_name);
    }


  if ( !DYNAMIC_DMA_BUFFER_SIZE )
    {
      fprintf( stderr, "Error: not using huge page tables\n");
      usage(prog_name);
    }
}


/*
*******************************************************************************
*
*                      initialization and termination 
*
*******************************************************************************
*/


void init_main( int argc, char** argv )
{
  init_dist_table();
  parse_input( argc, argv );
  task_init();
}


void finalize_main(void)
{
  task_end();
  parse_event_buffers();
  print_event_distributions();
}


void init_user_control_block( void ) 
{  
  user_cb.N = N;
  user_cb.dynamic_dma_buffer = (unsigned) dynamic_dma_buffer;
  user_cb.dynamic_dma_buffer_size = DYNAMIC_DMA_BUFFER_SIZE;
}

 
/*
*******************************************************************************
*
*                                 main
*
*******************************************************************************
*/


int main( int argc, char** argv ) 
{
  timer_t time1, time2, t_diff;
  struct timezone tz;
  double final_time;

  init_main( argc, argv );

  /* allocate the buffer that will contain the stt and the input text */
  allocate_huge_buffer();
  
  DB_DMA(printf("size of control block %d\n", sizeof( control_block_t ) );)

  /* initialize the user control block and pass the relevant
     information to the SPUs */
  init_user_control_block();

  printf("PPU: starting SPUs...\n");

  get_clock(time1);
  write_in_mbox_all( (unsigned) (&user_cb));

  finalize_main();

  get_clock(time2); 

  diff(final_time, time2, time1);

  printf( "Execution time\t: %2.6f seconds\n", final_time );


  return 0;
}

