/*
*******************************************************************************
*
* File:         runtime_ppu.c
* RCS:          $Header: $
* Description:  PPU run-time support 
* Author:       Fabrizio Petrini 
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/


#include "../common.h"
#include "runtime_ppu.h"
#include "dist.h"


/*
*******************************************************************************
*
*                             local defines
*
*******************************************************************************
*/


/* the initial value of the distribution size */
#define DISTRIBUTION_SIZE     8192

/* the threshold above which we switch to a queue implementation */
#define MAX_DISTRIBUTION_SIZE 8192

/* the number of SPUS */
unsigned N;

/* this is the pointer to the SPE code, used at thread creation time */
extern spe_program_handle_t main_spu;

#ifdef LIBSPE1

/* before spawning the threads, we create a group environment to
   contain it; this group id (gid) is the handle for that group */
spe_gid_t gid;

/* this is the handle which will be returned by spe_create_thread */
speid_t speids[MAX_SPU];

#else /* LIBSPE2 */

/* container for the thread data structures returned and used by
   spe_context_create and pthread_create */
typedef struct thread_data 
{
  spe_context_ptr_t speid;
  pthread_t pthread;
  void *argp;
} thread_data_t;

/* an element for each SPE */
thread_data_t spe_thread_data[MAX_SPU];

#endif /* LIBSPE1 */

/* this variable is used to return data regarding an abnormal return
   from the SPU */
int status[MAX_SPU];

/* MAX_SPU buffers of 256KB that mirror the LS in main memory */
char mirror_ls[MAX_SPU][LS_SIZE] align128;

/* arrays of large and small communication buffers */
unsigned int small_buffer[MAX_SPU][SMALL_BUFFER_SIZE] align16;
unsigned int large_buffer[MAX_SPU][LARGE_BUFFER_SIZE] align16;
communication_buffer in_small_buffer[MAX_SPU] align16;
communication_buffer in_large_buffer[MAX_SPU] align16;

/* we allocate one control block, that contains the info about the threads */
spe_control_block cb align128;

/* the prefix of the distribution files */
char dist_prefix[MAX_STRING] = "";

/* bi-dimensional array of distributions */
distribution* dist_table[MAX_SPU][MAX_EVENT];

/* each event is identified by an event id and an event name: this is
   an array of events, common.h */
distarray;


/*
*******************************************************************************
*
*                             event monitoring
*
*******************************************************************************
*/


/* initialize the distribution table: by default each table entry is
   null, and it is initialized only when there are events of the
   specific class generated by the SPU */
void init_dist_table()
{
  unsigned i, j;
  
  for( i = 0; i < N; i++ )
    {
      for( j = 0; j < MAX_EVENT; j++ )
	dist_table[i][j] = 0;
    }
}


void parse_event_buffers()
{
  unsigned int i, j, event_tag, event_val;

  /* scan the event log of each SPU */
  for( i = 0; i < (unsigned int) N; i++ )
    {
      if ( in_large_buffer[i].uival[0] > LARGE_BUFFER_SIZE )
        {
	  printf( "parse_event_buffers: spu %d in %d overflows buffer %d\n", 
		  i, in_large_buffer[i].uival[0], LARGE_BUFFER_SIZE );
	  exit(0);
        }
      
      /* extract different distributions for each class of events */
      for( j = 0; j < in_large_buffer[i].uival[0]; j++ )
        {
	  event_tag = ( large_buffer[i][j] >> EVENT_VALUE_BITS );
	  
	  if ( event_tag == OVERFLOW_EVENT )
            {
	      /* the tag and the actual value are stored in the
		 following two words */
	      event_tag = ( large_buffer[i][++j] >> EVENT_VALUE_BITS );
	      event_val = large_buffer[i][++j];
            }
	  else
	    event_val = ( large_buffer[i][j] & EVENT_VALUE_BITMASK );
	  
	  /* the first time we create the distribution */
	  if ( !dist_table[i][event_tag] )
            {
	      dist_table[i][event_tag] = 
		create_distribution( dist_array[event_tag].event_file_name, 
                            DISTRIBUTION_SIZE,
				     MAX_DISTRIBUTION_SIZE );
            }
	  
	  /* update the distribution */
	  update_distribution( dist_table[i][event_tag], event_val );
        }
    }
}


void print_event_distributions()
{
  unsigned i, j;
  char name[MAX_STRING];
  
  for( i = 0; i < N; i++ )
    for( j = 0; j < MAX_EVENT; j++ )
      if ( dist_table[i][j] )
	{
	  sprintf( name, "%s%s", dist_prefix, dist_array[j].event_file_name );
	  print_distribution( dist_table[i][j], name, i, 1.0 );
	  delete_distribution( dist_table[i][j] );
	}
}


/*
*******************************************************************************
*
*                        thread creation and termination
*
*******************************************************************************
*/


/* computes the ceiling of the base two logarithm of the given value */
unsigned int ceiling_log2 ( unsigned int value )
{
  unsigned int log = 0, bit_count = 0;
  
  while ( value )
    {
      if ( value & 0x1 )
	bit_count++;
      
      value = value >> 1;
      
      log++;
    }
  
  if ( bit_count > 1 )
    return log;
  else
    return log - 1;
}


#ifdef LIBSPE1
void create_group( int policy, int priority, int spe_events )
{
  /* create a group and double check that there are enough SPUS */
  gid = spe_create_group( policy, priority, spe_events );
  
  if ( !gid ) 
    {
      printf( "create_group: failed spe_create_group, errno %d\n", errno);
      exit(0);
    }
  
  if ( spe_group_max( gid ) <  (int) N ) 
    {
        printf( "create_group: only %d SPUs are available, requested %d\n", 
                spe_group_max( gid ), N );
        exit(0);
    }
  
  cb.spus = N;
  cb.log_spus = ceiling_log2( N );
}


void write_in_mbox( int spu_id, unsigned int val )
{
  while ( spe_write_in_mbox( speids[spu_id], val ) );
}


void write_in_mbox_all( unsigned int val )
{
  int i;
  
  /* start from the last one */
  for( i = N - 1; i >=0 ; i-- )
    while ( spe_write_in_mbox( speids[i], val ) );
}


unsigned int read_out_mbox( int spu_id )
{
  int mbox_reply;
  
  /* wait for the reply */
  while ( ( mbox_reply = spe_read_out_mbox( speids[spu_id] ) ) == -1 );
  
  return (unsigned int) mbox_reply;
}


void initialize_threads()
{ 
  unsigned i, j;
  
  /* send a start message to all threads: the threads will issue
     independent gets to copy the control block from main memory */
  for( i = 0; i < N; i++ )
    write_in_mbox( i, i );
  
  for( i = 0; i < N; i++ )
    {
      j = read_out_mbox( i );
      
      if ( j != i )
	printf( "initialize_threads: returned value %d and id %d differ\n",
		j, i );
    }
  
  /* and write back into the mailboxes again, to make sure that there
     is a barrier synchronization */
  for( i = 0; i < N; i++ )
    write_in_mbox( i, i ); 
}


void task_init()
{
  unsigned i;
  DB(unsigned j;)
    
  /* create the thread group */
  create_group( SCHED_OTHER, 0, 1 );
  
  /* allocate the SPUS: we force processor affinity */
  for( i = 0; i < N; i++ )
    {
      speids[i] = spe_create_thread( gid, &main_spu, 
				     (unsigned long long*) &cb, 
				     NULL, 1<<i, SPE_MAP_PS );
      if ( !speids[i] ) 
        {
	  printf( "create_threads: failed to create thread %d, errno %d\n", 
		  i, errno );
	  exit (0);
        }
    }
  
  for( i = 0; i < N; i++ )
    {
      if ( !( cb.ls[i] = (unsigned int) spe_get_ls( speids[i] ) ) )
        {
            printf( "create_thread: failed spe_get_ls %d, errno %d\n", 
		    i, errno);
            exit (0);
        }

      /*
        if ( !( cb.ps_cntl[i] = 
	(unsigned int) spe_get_ps_area( speids[i], SPE_CONTROL_AREA ) ) )
        {
	printf( "create_thread: failed spe_get_ps_area %d, errno %d\n", 
	i, errno);
	exit (0);
        }
      */
      
      /* initialize the array input counters */
      in_small_buffer[i].uival[0] = 0;
      in_large_buffer[i].uival[0] = 0;
      
      /* set the array pointers */
      cb.mirror_ls[i] = (unsigned int) mirror_ls[i];
      cb.small_buffer[i] = (unsigned int) small_buffer[i];
      cb.in_small_buffer[i] = (unsigned int) &in_small_buffer[i];
      cb.large_buffer[i] = (unsigned int) large_buffer[i];
      cb.in_large_buffer[i] = (unsigned int) &in_large_buffer[i];
    }
  
  /* tag the ls buffers */
  DB(for( i = 0; i < MAX_SPU; i++ )
       for( j = 0; j < LS_SIZE; j++ )
	 mirror_ls[i][j] = (char) i;)
    
    
  /* double check that the control block size is correct: compiler bug? */   
  if ( CB_SIZE != sizeof( spe_control_block ) )
    {
      printf( "task_init: mismatch in control block size: %u actual %u\n",
	      CB_SIZE, sizeof( spe_control_block ) );
      exit(0);
    }
  
  /* handshake with the spu threads */  
  initialize_threads();  
}


void wait_for_spu( int i )
{
  /* wait for the SPU to complete */
  if ( spe_wait( speids[i], &status[i], 0 ) )
    {
      printf( "wait_for_spu: failed spe_wait thread %d status %d errno %d\n", 
	      i, status[i], errno);
      exit (0);
    }
}


void task_end()
{
  unsigned i;
  
  /* wait for the SPUS to complete */
  for( i = 0; i < N; i++ )
    {
      wait_for_spu( i );
      printf("PPU: completed spu %d\n", i );
    }
  
  /* Issue a sync, just to be safe. */
  __asm__ __volatile__ ("sync" : : : "memory");
  
  /* print termination message */
  DB(printf( "PPU successfully terminating\n" );)
}


#else /* LIBSPE2 */


void *pthread_function( void *arg ) 
{
  thread_data_t *datap = (thread_data_t *) arg;
  unsigned entry = SPE_DEFAULT_ENTRY;
  int rc;

  if ( ( rc = spe_context_run( datap->speid, &entry, 0, datap->argp, 
			       NULL, NULL ) ) < 0 ) 
    {
      printf( "pthread_function: failed spe_context_run"
	      "( rc = %d, errno = %d, strerror = %s )\n", 
	      rc, errno, strerror( errno ) );
      exit(0);
    }

  pthread_exit(NULL);
}


void write_in_mbox( int spu_id, unsigned int val )
{
  while ( spe_in_mbox_write( spe_thread_data[spu_id].speid, &val, 1, 
			     SPE_MBOX_ALL_BLOCKING ) == 0 );
}


void write_in_mbox_all( unsigned int val )
{
  int i;
  
  /* start from the last one */
  for( i = N - 1; i >=0 ; i-- )
    while ( spe_in_mbox_write( spe_thread_data[i].speid, &val, 1, 
			       SPE_MBOX_ALL_BLOCKING ) == 0 );
}


unsigned int read_out_mbox( int spu_id )
{
  unsigned int mbox_reply;
  
  while ( spe_out_mbox_read( spe_thread_data[spu_id].speid, &mbox_reply, 1 ) ==
	  0 );
  
  return mbox_reply;
}


void initialize_threads()
{ 
  unsigned i, j;
  
  /* send a start message to all threads: the threads will issue
     independent gets to copy the control block from main memory */
  for( i = 0; i < N; i++ )
    write_in_mbox( i, i );
  
  for( i = 0; i < N; i++ )
    {
      j = read_out_mbox( i );
      
      if ( j != i )
	printf( "initialize_threads: returned value %d and id %d differ\n",
		j, i );
    }
  
  /* and write back into the mailboxes again, to make sure that there
     is a barrier synchronization */
  for( i = 0; i < N; i++ )
    write_in_mbox( i, i ); 
}


void task_init()
{
  unsigned i;
  DB(unsigned j;)
  int rc;

  cb.spus = N;
  cb.log_spus = ceiling_log2( N );

  for( i = 0; i < N; i++ )
    {
      /* create context */
      if ( ( spe_thread_data[i].speid = spe_context_create( 0, NULL ) ) == 
	   NULL )
        {
          printf( "task_init: failed spe_context_create"
		  "( errno = %d strerror = %s )\n", errno, strerror( errno ) );
          exit(0);
        }

      /* load the program */
      if ( ( rc = spe_program_load( spe_thread_data[i].speid, 
				    &main_spu ) ) != 0 )
        {
          printf( "task_init: failed spe_program_load"
		  "( errno = %d strerror = %s )\n", errno, strerror( errno ) );
          exit(0);
        }

      /* initialize the data */
      spe_thread_data[i].argp = (unsigned long long *) &cb;

      /* create the SPE thread */
      if ( ( rc = pthread_create( &spe_thread_data[i].pthread, NULL, 
				  &pthread_function, 
				  &spe_thread_data[i] ) ) != 0 )
        {
          printf( "task init: failed pthread_create"
		  "( errno = %d strerror = %s )\n", errno, strerror( errno ) );
          exit(0);
        }
    }


  for( i = 0; i < N; i++ )
    {
      if ( !( cb.ls[i] = 
	      (unsigned) spe_ls_area_get( spe_thread_data[i].speid ) ) || 
	   cb.ls[i] == 0xffffffff )
        {
            printf( "create_thread: failed spe_get_ls %d, errno %d, ls %p\n",
		    i, errno, (void*) cb.ls[i] );
            exit (0);
        }

      /*
        if ( !( cb.ps_cntl[i] = 
	(unsigned int) spe_get_ps_area( speids[i], SPE_CONTROL_AREA ) ) )
        {
	printf( "create_thread: failed spe_get_ps_area %d, errno %d\n", 
	i, errno);
	exit (0);
        }
      */
      
      /* initialize the array input counters */
      in_small_buffer[i].uival[0] = 0;
      in_large_buffer[i].uival[0] = 0;
      
      /* set the array pointers */
      cb.mirror_ls[i] = (unsigned int) mirror_ls[i];
      cb.small_buffer[i] = (unsigned int) small_buffer[i];
      cb.in_small_buffer[i] = (unsigned int) &in_small_buffer[i];
      cb.large_buffer[i] = (unsigned int) large_buffer[i];
      cb.in_large_buffer[i] = (unsigned int) &in_large_buffer[i];
    }
  
  /* tag the ls buffers */
  DB(for( i = 0; i < MAX_SPU; i++ )
       for( j = 0; j < LS_SIZE; j++ )
	 mirror_ls[i][j] = (char) i;)
    
    
  /* double check that the control block size is correct: compiler bug? */   
  if ( CB_SIZE != sizeof( spe_control_block ) )
    {
      printf( "task_init: mismatch in control block size: %u actual %u\n",
	      CB_SIZE, sizeof( spe_control_block ) );
      exit(0);
    }
  
  /* handshake with the spu threads */  
  initialize_threads();  
}


void task_end()
{
  unsigned i;
  int rc;

  /* wait for SPEs to all finish */
  for( i = 0;  i < N; i++ ) 
    {
      /* join thread */
      if ( ( rc = pthread_join( spe_thread_data[i].pthread, NULL ) ) != 0 )
        {
          printf( "task_end: failed pthread_join"
		  "( rc = %d, errno = %d strerror = %s)\n", 
		  rc, errno, strerror( errno ) );
          exit(0);
        }

      /* destroy the context */
      if ( ( rc = spe_context_destroy( spe_thread_data[i].speid ) ) != 0 )
        {
          printf( "task_end: failed spe_context_destroy"
		  "( rc = %d, errno = %d strerror = %s )\n", 
		  rc, errno, strerror( errno ) );
          exit(0);
        }
    }
  
  /* Issue a sync, just to be safe. */
  __asm__ __volatile__ ("sync" : : : "memory");
  
  /* print termination message */
  DB(printf( "PPU successfully terminating\n" );)
}


#endif



