/*
*******************************************************************************
*
* File:         runtime_spu.c
* RCS:          $Header: $
* Description:  The SPU communication runtime support
* Author:       Fabrizio Petrini 
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/



#include "runtime_spu.h"



/*
*******************************************************************************
*
*                              initialization
*
*******************************************************************************
*/


/* the spu id and the number of spus: initialized by
   initialize_spu_thread */
unsigned int my_id align128;
unsigned int spus;

/* the ceiling of the base 2 logarithm of the number of spus: the
   number of steps in a reduce tree or a broadcast in a collective
   communication */
unsigned int log_spus;

/* the local copy of the control block, to be filled by the initial DMA */
spe_control_block cb align128;


/*
*******************************************************************************
*
*                                monitoring 
*
*******************************************************************************
*/


/* local and remote in pointers for the small and large buffers */
unsigned int in_small = 0, in_large = 0; 
unsigned int remote_in_small align128 = 0;
unsigned int remote_in_large align128 = 0;

/* small and large buffers */
unsigned int spu_small_buffer[SPU_SMALL_BUFFER_SIZE] align128;
unsigned int spu_large_buffer[SPU_LARGE_BUFFER_SIZE] align128;


/*
*******************************************************************************
*
*                       communication and synchronization 
*
*******************************************************************************
*/

#ifdef COMMUNICATION_RUNTIME

/* in order to avoid alignment problems we use an entire 16-byte EIB
   transaction unit for the synchronization variables */
volatile communication_buffer receiver_ack[MAX_SPU] align128;
volatile communication_buffer send_buffer_addr[MAX_SPU] align128;


/* the reduce send and receive buffers: the input buffer is volatile,
   so we can check the presence of an incoming message; we use an
   array of buffers to avoid race conditions: each sender can put its
   message into a distinct receive buffer */
volatile communication_buffer reduce_recv_buffer[MAX_SPU] align128;
communication_buffer reduce_send_buffer align128;

/* the phase counters used by the reduce and broadcast synchronization
   trees */
int sync_reduce_phase;
int sync_broadcast_phase;


/* the binary tree table: rather than having a mathematical definition
   of the binary tree, we use a table format, to fine-tune the
   communication pattern of the reduce tree (and, symmetrically, the
   broadcast tree), and deal with specific issues in the mapping of
   the communication pattern on the EIB; the following table works for
   a 16-SPU system: there is a row for each SPU and a column for each
   phase (4 phases = log_2 (16)); 16, which is supposed to be equal to
   MAX_SPU, means nop, a positive number (different from 16) means we
   send to a partner in the given reduce phase while a negative number
   identifies an incoming message; for example, SPU 0 receives from
   SPUs 2, 4, 1 and 8 */
int binary_tree[16][4] = 
  {
    /* SPU  0 */ { -2,  -4,  -1, -8},
    /* SPU  1 */ { -3,  -5,   0, 16},
    /* SPU  2 */ {  0,  16,  16, 16},
    /* SPU  3 */ {  1,  16,  16, 16},
    /* SPU  4 */ { -6,   0,  16, 16},
    /* SPU  5 */ { -7,   1,  16, 16},
    /* SPU  6 */ {  4,  16,  16, 16},
    /* SPU  7 */ {  5,  16,  16, 16}, 
    /* SPU  8 */ {-10, -12,  -9,  0},
    /* SPU  9 */ {-11, -13,   8, 16},
    /* SPU 10 */ {  8,  16,  16, 16},
    /* SPU 11 */ {  9,  16,  16, 16},
    /* SPU 12 */ {-14,   8,  16, 16},
    /* SPU 13 */ {-15,   9,  16, 16},
    /* SPU 14 */ { 12,  16,  16, 16},
    /* SPU 15 */ { 13,  16,  16, 16}
  };

/* the binomial tree table: rather than having a mathematical
   definition of the binomial tree, we use a table format, to
   fine-tune the communication pattern of the reduce tree (and,
   symmetrically, the broadcast tree), and deal with specific issues
   in the mapping of the communication pattern on the EIB; the
   following table works for a 16-SPU system: there is a row for each
   SPU and a column for each phase (4 phases = log_2 (16)); 16, which
   is supposed to be equal to MAX_SPU, means nop, a positive number
   (different from 16) means we send to a partner in the given reduce
   phase while a negative number identifies an incoming message; for
   example, SPU 0 receives from SPUs 2, 4, 1 and 8 
*/
int binomial_tree[16][4] = 
  {
    /* SPU  0 */ { -2,  -4,  -1, -8},
    /* SPU  1 */ { -3,  -5,   0, 16},
    /* SPU  2 */ {  0,  16,  16, 16},
    /* SPU  3 */ {  1,  16,  16, 16},
    /* SPU  4 */ { -6,   0,  16, 16},
    /* SPU  5 */ { -7,   1,  16, 16},
    /* SPU  6 */ {  4,  16,  16, 16},
    /* SPU  7 */ {  5,  16,  16, 16}, 
    /* SPU  8 */ {-10, -12,  -9,  0},
    /* SPU  9 */ {-11, -13,   8, 16},
    /* SPU 10 */ {  8,  16,  16, 16},
    /* SPU 11 */ {  9,  16,  16, 16},
    /* SPU 12 */ {-14,   8,  16, 16},
    /* SPU 13 */ {-15,   9,  16, 16},
    /* SPU 14 */ { 12,  16,  16, 16},
    /* SPU 15 */ { 13,  16,  16, 16}
  };


int binomial_tree1[16][4] = 
  {
    /* SPU  0 */ { -1,  -2,  -4, -8},
    /* SPU  1 */ {  0,  16,  16, 16},
    /* SPU  2 */ { -3,   0,  16, 16},
    /* SPU  3 */ {  2,  16,  16, 16},
    /* SPU  4 */ { -5,  -6,   0, 16},
    /* SPU  5 */ {  4,  16,  16, 16},
    /* SPU  6 */ { -7,   4,  16, 16},
    /* SPU  7 */ {  6,  16,  16, 16}, 
    /* SPU  8 */ { -9, -10, -12,  0},
    /* SPU  9 */ {  8,  16,  16, 16},
    /* SPU 10 */ {-11,   8,  16, 16},
    /* SPU 11 */ { 10,  16,  16, 16},
    /* SPU 12 */ {-13, -14,   8, 16},
    /* SPU 13 */ { 12,  16,  16, 16},
    /* SPU 14 */ {-15,  12,  16, 16},
    /* SPU 15 */ { 14,  16,  16, 16}
  };



/* precomputed data structures for the lightweight version of the
   barrier synchronization */
vector unsigned int reduce_cond1;
vector unsigned int reduce_cond2;
vector unsigned int broadcast_cond1;
vector unsigned int broadcast_cond2;

dma_list_element reduce_dma[4] align128;
dma_list_element broadcast_dma[4] align128;

int light_phases = 0;

#else

/* the send and receive buffers: the receive buffer is volatile, so we
   can check the presence of an incoming message; we use an array of
   buffers to avoid race conditions: each sender can put its message
   into a distinct receive buffer */
volatile communication_buffer recv_buffer[MAX_SPU] align128;
communication_buffer send_buffer align128;

#endif /* COMMUNICATION_RUNTIME */

/*
*******************************************************************************
*
*                               utilities
*
*******************************************************************************
*/


void *spe_memset( void *dest, int val, size_t len )
{
  register unsigned char *ptr = (unsigned char*) dest;
  while (len-- > 0)
    *ptr++ = val;
  return dest;
}


#ifdef COMMUNICATION_RUNTIME


unsigned log_2 ( unsigned value )
{
  unsigned int log = -1;
  
  while ( value )
    {
      value = value >> 1; 
      log++;
    }
  
  return log;
}


unsigned int_log2( unsigned value ) 
{
  unsigned last_count = 0, count = 0;

  while ( value ) 
    {
      if ( value & 0x1 )
	last_count = count;
      value = value >> 1;
      count ++;
    }
  return last_count; 
} 


unsigned ceil_power_of_two( unsigned value ) 
{
  unsigned last_count = 0, count = 0;
  unsigned tmp_value = value, out = 1;
  
  while ( tmp_value ) 
    {
      if ( tmp_value & 0x1 )	
	last_count = count;
      
      tmp_value = tmp_value >> 1;
      count ++;
    }
  
  for( ; last_count > 0; last_count--)
    out = out << 1;
  
  if(value > out) 
    return out << 1;
  else 
    return out;
}


/*
*******************************************************************************
*
*                                  dmas
*
*******************************************************************************
*/


/* print the put and get parameters */
void print_put( unsigned ls, unsigned ea, unsigned size, unsigned tag )
{
  spe_printf( "put ls 0x%x (%u), ea 0x%x (%u), size %u tag %u\n", 
	      ls, ls, ea, ea, size, tag );
}


void print_get( unsigned ls, unsigned ea, unsigned size, unsigned tag )
{
  spe_printf( "get ls 0x%x (%u), ea 0x%x (%u), size %u tag %u\n", 
	      ls, ls, ea, ea, size, tag );
}


/* check alignment on a given boundary, which must be a power of 2 */
unsigned int misaligned( uint64_t addr, unsigned int boundary )
{
  return (unsigned int) ( addr % boundary );
}


/* The MFC supports naturally aligned transfer sizes of 1, 2, 4 or 8
   bytes, and multiple of 16 bytes, with a maximum transfer size of
   16KB; the tricky part is that "naturally aligned" means that source
   and destination must have the same address in a 16-byte word */
DB_DMA(void check_dma_align( void* ls, uint64_t rs, unsigned int size )
{
  unsigned int align_size = ( size > 16 || size == 0 ) ? 16: size;

  if( (uint32_t) ls == 0 || (uint32_t ) rs == 0 )
    {
      spe_printf( "check_dma_align: null address local 0x%x remote 0x%x\n",
		  (uint32_t) ls, (uint32_t) rs );
      while (1);
    }

  if ( (((uint32_t) ls) & 0xF) != (((uint32_t) rs) & 0xF) )
    {
      spe_printf( "check_dma_align: dma not naturally aligned 0x%x 0x%llx\n",
		  (uint32_t) ls, rs );
      while (1);
    }

  if ( size > MAX_DMA_SIZE )
    {
      spe_printf( "check_dma_align: exceeding maximum DMA size %u\n", size );
      while (1);
    }

  if ( size > 16 && size % 16 )
    {
      spe_printf( "check_dma_align: wrong size %u\n", size );
      while (1);
    }
      
  if ( size == 0 || size == 1 || size == 2 || size == 4 || size == 8 || 
       size == 16 || align_size == 16 )
    {
      if ( misaligned( (uint32_t) ls, align_size ) )
	{
	  spe_printf( "check_dma_align: ls address 0x%x, size %u aligned %u\n",
		      (uint32_t) ls, size, align_size );
	  while (1);
	}

      if ( misaligned( rs, align_size ) )
	{
	  spe_printf( "check_dma_align: ea address 0x%llx, "
		      "size %u aligned %u\n", rs, size, align_size );
	        while (1);
	}
    }
  else
    {
      spe_printf( "check_dma_align: wrong size %u\n", size );
      while (1);
    }
})


void blocking_put( void* ls, uint64_t rs, unsigned int size, unsigned int tag )
{
  DB_DMA(check_dma_align( ls, rs, size );)

  DB_DMA(spe_printf( "spu %u: starting put 0x%x 0x%llx %u \n", my_id, 
		     (unsigned int) ls, (unsigned long long int) rs, size);)

  DB_DMA(print_put( (unsigned) ls, (unsigned) rs, size, tag );)	

  /* the first parameter is the address in local store to place the data */
  /* the second parameter holds the main memory address                  */
  /* the third parameter holds the number of bytes to DMA                */
  /* the fourth parameter identifies a "tag" to associate with this DMA  */
  /* (this should be a number between 0 and 31, inclusive)               */
  /* the last two parameters are only useful if you've implemented your  */
  /* own cache replacement management policy.  Otherwise set them to 0.  */
  mfc_put( ls, rs, size, tag, 0, 0);

  DB_DMA(spe_printf( "spu %u: put issued 0x%x 0x%llx %u\n", my_id, 
		     (unsigned int) ls, (unsigned long long int) rs, size );)
    
  /* now, we set the "tag bit" into the correct channel on the hardware  */
  /* this is always 1 left-shifted by the tag specified with the DMA     */
  /* for whose completion you wish to wait.                              */
  mfc_write_tag_mask( 1 << tag );

  DB_DMA(spe_printf( "spu %u: put after write_tag_mask\n", my_id );)
  
  /* now, issue the read and wait to guarantee DMA completion */
  mfc_read_tag_status_all();

  DB_DMA(spe_printf( "spu %u: put completed\n", my_id );)
}


/* for dmas larger than 16KB */
void large_blocking_put( void* ls, uint64_t rs, unsigned int size, 
			 unsigned int tag )
{
  unsigned dma_size_left = size, local_buffer = (unsigned) ls,
    remote_buffer = (unsigned) rs, dma_size;

  while ( dma_size_left )
    {
      dma_size = ( dma_size_left > MAX_DMA_SIZE ) ? MAX_DMA_SIZE: 
	dma_size_left;
      
      blocking_put( (void*) local_buffer, remote_buffer, dma_size, tag );

      local_buffer += dma_size;
      remote_buffer += dma_size;
      dma_size_left -= dma_size;
    }
}


void blocking_get( void* ls, uint64_t rs, unsigned int size, unsigned int tag )
{
  DB_DMA(check_dma_align( ls, rs, size );)

  DB_DMA(spe_printf( "spu %u: starting get\n", my_id );)
  DB_DMA(print_get( (unsigned) ls, (unsigned) rs, size, tag );)	

  mfc_get( ls, rs, size, tag, 0, 0);

  DB_DMA(spe_printf( "spu %u: get issued 0x%x 0x%llx %u\n", my_id, 
		     (unsigned int) ls, (unsigned long long int) rs, size );)

  mfc_write_tag_mask( 1 << tag );

  DB_DMA(spe_printf( "spu %u: get after write_tag_mask\n", my_id );)

  mfc_read_tag_status_all();

  DB_DMA(spe_printf( "spu %u: get completed\n", my_id );)
}


/* for dmas larger than 16KB */
void large_blocking_get( void* ls, uint64_t rs, unsigned int size, 
			 unsigned int tag )
{
  unsigned dma_size_left = size, local_buffer = (unsigned) ls,
    remote_buffer = (unsigned) rs, dma_size;

  while ( dma_size_left )
    {
      dma_size = ( dma_size_left > MAX_DMA_SIZE ) ? MAX_DMA_SIZE: 
	dma_size_left;
      
      blocking_get( (void*) local_buffer, remote_buffer, dma_size, tag );

      local_buffer += dma_size;
      remote_buffer += dma_size;
      dma_size_left -= dma_size;
    }
}


/* write into the remote mailbox; to be used with extreme care,
   because this function does not check for mailbox overflows and that
   the actual mailbox write has been completed successfully; it is the
   programmer's responsibility to verify that everything is OK */
void write_in_mbox( unsigned int partner, unsigned int value, 
		    unsigned int tag )
{
  uint64_t remote = cb.ps_cntl[partner] + 0xC;
  uint32_t putbuf[4] align16;

  /* write into remote mailbox */
  putbuf[3] = value;
  blocking_put( (void*) &putbuf[3], remote, sizeof(uint32_t), tag );
}


/*
*******************************************************************************
*
*                                 waits
*
*******************************************************************************
*/


void wait (unsigned int tag) 
{
  mfc_write_tag_mask( 1 << tag );
  mfc_read_tag_status_all();
}


/* this is a non-blocking call: it only checks the presence of tags */
unsigned wait_immediate( unsigned wait_mask ) 
{
  unsigned completion_mask;

  /* setting write tag-group mask */
  mfc_write_tag_mask(wait_mask); 

  mfc_write_tag_update_immediate();
  completion_mask = mfc_read_tag_status();

  return completion_mask;
}


/* and this is a blocking call: it waits for any message in the tag
   classes to arrive  */
unsigned wait_any( unsigned wait_mask ) 
{
  unsigned completion_mask;

  /* setting write tag-group mask */  
  mfc_write_tag_mask(wait_mask);

  /* setting the tag status update policy */
  mfc_write_tag_update_any(); 
  completion_mask = mfc_read_tag_status();

  return completion_mask;
}


boolean_t tag_is_outstanding( unsigned dma_tag )
{
  unsigned wait_mask = 1 << dma_tag;
  
  if ( wait_immediate(wait_mask) == 0 )
    /* if wait_immediate returns 0, 
       then the tag is outstanding
     */
    return True;
  else 
    return False;
}


/*
*******************************************************************************
*
*                       point to point communication
*
*******************************************************************************
*/


void init_point_to_point()
{
  int i;

  for( i = 0; i < MAX_SPU; i++ )
    {
      receiver_ack[i].uival[0] = 0;
      send_buffer_addr[i].uival[0] = LS_SIZE;
    }
}


/* the blocking send takes the receiver id, which is another SPE, the
   sender buffer, which is in local storage, and the message size */
void blocking_send( unsigned int receiver, void* send_buffer, 
		    unsigned int size, unsigned int tag )
{
  DB_PTP(int value;)
  communication_buffer buf align16;
    
  DB_PTP(if ( receiver > spus )
    spe_printf( "blocking_send: spu %u, receiver is %u, spus is %u\n", 
		my_id, receiver, spus );

  if ( (uint32_t) send_buffer + size >= LS_SIZE )
    spe_printf( "blocking_send: spu %u, buffer 0x%u size %u is "
		"out of range\n", my_id, (uint32_t) send_buffer, size );

  if ( size > MAX_DMA_SIZE )
    spe_printf( "blocking_send: spu %u, dma size %u is too large\n", 
		my_id, size );)
	 
  /* the blocking protocol works in the following way; 

  there are two main data structures,

  (1) send_buffer_addr, a buffer
  where each potential sender writes the local send_address (an LS
  address, that must be translated to an effective address using the
  info in the spe_control_block) and the message size, and

  (2) a synchronization buffer, receiver_ack, used by the receiver to
  inform the sender that the message has been copied into the final
  destination by the receiver */

  DB_PTP(spe_printf( "spu %u sending to %u\n", my_id, receiver );)

  /* in order to avoid a nasty bug we bracket the synchronization put
     with a mailbox write */  
  DB_SYNC(write_in_mbox( receiver, my_id );)
  
  /* write the address of the source buffer and the message size to
     the destination channel */
  buf.uival[0] = (unsigned int) send_buffer;
  buf.uival[1] = size; 
  blocking_put( (void*) &buf, cb.ls[receiver] + 
		(uint32_t) &send_buffer_addr[my_id], 
		sizeof( communication_buffer ), tag );

  DB_PTP(spe_printf( "spu %u sending to %u, after blocking_put\n", 
		     my_id, receiver );)

  /* we also bracket the loop on the counter, to make sure that we
     avoid the read-write bug */  
  DB_SYNC(value = spu_readch( SPU_RdInMbox );
 
  if ( value != receiver )
    spe_printf( "spu %u: mailbox mismatch, actual %u, expected %u\n", 
		value, receiver );)

  /* let's wait from the ack from the receiver */
  while ( receiver_ack[receiver].uival[0] == 0 );

  DB_PTP(spe_printf( "spu %u sending to %u, after waiting for ack\n", 
		     my_id, receiver );)

  /* and reset the flag */
  receiver_ack[receiver].uival[0] = 0;
}


/* the blocking_recv takes the sender id, which is another SPE, the
   receive buffer, which is in local storage, and the message size */
void blocking_recv( unsigned int sender, void* recv_buffer, 
		    unsigned int size, unsigned int tag )
{
  unsigned int ack align16 = 0xFFFFFFFF;
  DB_PTP(unsigned int value;)

  DB_PTP(if ( sender > spus )
    spe_printf( "blocking_recv: spu %u, sender is %u, spus is %u\n", 
		my_id, sender, spus );

  if ( (uint32_t) recv_buffer + size >= LS_SIZE )
    spe_printf( "blocking_recv: spu %u, buffer 0x%u size %u "
		"is out of range\n", my_id, (uint32_t) recv_buffer, size );

  if ( size > MAX_DMA_SIZE )
    spe_printf( "blocking_recv: spu %u, dma size %u is too large\n", 
		my_id, size );)

  /* we bracket the loop on the counter, to make sure that we avoid
     the read-write bug */  
  DB_SYNC(value = spu_readch( SPU_RdInMbox );
 
  if ( value != sender )
    spe_printf( "spu %u: mailbox mismatch, actual %u, expected %u\n", 
		value, sender );)

  /* wait for the source address from the receiver */
  while ( send_buffer_addr[sender].uival[0] == LS_SIZE );

  if ( __builtin_expect( ( send_buffer_addr[sender].uival[1] > size ), 0 ) )
    {
      spe_printf( "spu %d, blocking_recv: buffer overflow size %u, "
		  "available %u\n", my_id,  send_buffer_addr[sender].uival[1], 
		  size );
    }

  DB_PTP(spe_printf( "spu %u receiving from %u, after receiving addr 0x%x\n",
		     my_id, sender, send_buffer_addr[sender].uival[0] );)

  /* in order to avoid a nasty bug we bracket the synchronization put
     with a mailbox write */  
  DB_SYNC(write_in_mbox( sender, my_id );)

  /* get the message into the receive buffer */
  blocking_get( recv_buffer, cb.ls[sender] + send_buffer_addr[sender].uival[0],
		size, tag );

  DB_PTP(spe_printf( "spu %u receiving from %u, after blocking_get\n", 
		     my_id, sender );)

  /* reset the buffer address */
  send_buffer_addr[sender].uival[0] = LS_SIZE;

  /* send an ack back to the sender */
  blocking_put( (void*) &ack, cb.ls[sender] + (uint32_t) &receiver_ack[my_id],
		sizeof(uint32_t), tag );

  DB_PTP(spe_printf( "spu %u receiving from %u, after blocking_put\n", 
		     my_id, sender );)
}


/*
*******************************************************************************
*
*                        collective communication
*
*******************************************************************************
*/


void init_collectives()
{
  int partner, phase, tmp;
  unsigned int cond1, cond2;

  /* initialize the reduce communication buffers */
  spe_memset( (void*) &reduce_send_buffer, 0, sizeof( communication_buffer ) );
  spe_memset( (void*) &reduce_recv_buffer[0], 0, 
	      MAX_SPU * sizeof( communication_buffer ) );

  /* phase counters */
  sync_reduce_phase = 0;
  sync_broadcast_phase = 3;

  /* set the number of phases */
  tmp = spus >> 1;

  while ( tmp )
    {
      light_phases++;
      tmp = tmp >> 1;
    }

  light_phases += ( ( 1 << light_phases ) == (int) spus ) ? 0: 1;

  /* initialize the vector conditions */
  
  /* reduce: we visit all four colums of the table */
  for( phase = 0; phase < light_phases; phase++ )
    {
      partner = binomial_tree1[my_id][phase];

      cond1 = (unsigned int) ( partner >= 0 && partner < (int) spus );
      reduce_cond1 = spu_insert( cond1, reduce_cond1, phase ); 
      cond2 = (unsigned int) ( -((int) spus) < partner && partner < 0 );
      reduce_cond2 = spu_insert( cond2, reduce_cond2, phase ); 

      /* initialize the dma list */
      reduce_dma[phase].ea_low =  cb.ls[partner] + 
	((unsigned int) &reduce_recv_buffer[my_id] );
      reduce_dma[phase].size.all32 = sizeof( communication_buffer );
    }

  /* broadcast: the tree is visited in the reverse direction */
  for( phase = light_phases -1; phase >= 0; phase-- )
    {
      partner = binomial_tree1[my_id][phase];

      cond1 = (unsigned int) ( ((int) -spus) < partner && partner < 0 ); 
      broadcast_cond1 = spu_insert( cond1, broadcast_cond1, phase ); 
      cond2 = (unsigned int) ( partner >= 0 && partner < (int) spus );
      broadcast_cond2 = spu_insert( cond2, broadcast_cond2, phase ); 

      /* initialize the dma lists */
      broadcast_dma[phase].ea_low =  cb.ls[-partner] + 
	((unsigned int) &reduce_recv_buffer[my_id] );
      broadcast_dma[phase].size.all32 = sizeof( communication_buffer );
    }
}


/* sync_reduce executes the reduce phase of a synchronization algorithm
   that allows overlapping of communication and computation;
   sync_reduce returns FALSE until it traverses the synchronization
   tree, and must be called until it returns TRUE, which means that
   the reduce has been completed; the variable sync_reduce_phase keeps
   track of the current stage in the reduction tree */
int sync_reduce( unsigned int tag )
{
  int partner;

  /* MAX_SPU, which is expected to be 16, identifies an idle
     sync_reduce_phase; a non-negative number means that the SPU sends
     to a partner while a negative one means that we receive from a
     partner */
  partner = binary_tree[my_id][sync_reduce_phase];

  if ( partner >= 0 && partner < (int) spus )
    {
      /* set the presence byte */
      reduce_send_buffer.cval[15] = 1;
   
      /* send the value */
      blocking_put( (void*) &reduce_send_buffer, 
		    cb.ls[partner] + ((unsigned) &reduce_recv_buffer[my_id] ), 
		    sizeof( communication_buffer ), 
		    tag );

      /* increment the phase counter */
      sync_reduce_phase++;
    }
  else if ( -((int) spus) < partner && partner < 0 )
    {
      volatile communication_buffer* rbuf = 
	(communication_buffer*) &reduce_recv_buffer[-partner];
	
      /* check for message arrival */
      if ( rbuf->cval[15] )
	{
	  /* reset the buffer */
	  rbuf->cval[15] = 0;

	  /* increment the phase counter */
	  sync_reduce_phase++;
	}
    }
  else
    /* increment the phase counter: partner is out of range */
    sync_reduce_phase++;

  if ( sync_reduce_phase == 4 )
    {
      /* reset the phase counter */
      sync_reduce_phase = 0;
      return TRUE;
    }
  else
    return FALSE;
}


/* symmetric function of sync_reduce */
int sync_broadcast( unsigned int tag )
{
  int partner;

  /* MAX_SPU identifies an idle phase; a negative number means that
    the SPU sends to a partner while a positive one means that we
    receive from a partner */
  partner = binary_tree[my_id][sync_broadcast_phase];

  if ( ((int) -spus) < partner && partner < 0 )
    {
      /* set the presence byte */
      reduce_send_buffer.cval[15] = 1;

      /* send the value */
      blocking_put( (void*) &reduce_send_buffer, 
		    cb.ls[-partner] + 
		    ((unsigned) &reduce_recv_buffer[my_id]), 
		    sizeof( communication_buffer ), 
		    tag );

      /* decrement the counter */
      sync_broadcast_phase--;
    }
  else if ( partner >= 0 && partner < (int) spus )
    {
      volatile communication_buffer* rbuf = 
	(communication_buffer*) &reduce_recv_buffer[partner];
      
      /* check for message arrival */
      if ( rbuf->cval[15] )
	{
	  /* reset the buffer */
	  rbuf->cval[15] = 0;

	  /* decrement the counter */
	  sync_broadcast_phase--;
	}
    }
  else
    /* increment the phase counter: partner is out of range */
    sync_broadcast_phase--;

  if ( sync_broadcast_phase < 0 )
    {
      /* reset the phase counter */
      sync_broadcast_phase = 3;
      return TRUE;
    }
  else
    return FALSE;
}

/* the all-reduce is a two-phase operation, that first collects an
   associative value on SPE 0 using a binomial reduce tree, to then
   distribute back the result using the same binomial tree to all the
   members of the all_reduce collective communication */

void all_reduce( communication_buffer* buffer, 
		 const reduce_op_type op, 
		 const unsigned int tag )
{
  /* the number of phases is the ceiling of the log of the number of
     SPUs */
  int phase, partner;

  /* initialize the local communication buffer */
  switch ( op )
    {
    case REDUCE_BARRIER:
      break;
    case REDUCE_INT:
    case REDUCE_INT_MAX:
      reduce_send_buffer.ival[0] = buffer->ival[0];
      break;
    case REDUCE_DOUBLE:
    case REDUCE_DOUBLE_MAX:
      reduce_send_buffer.dval[0] = buffer->dval[0];
      break;
    default:
      spe_printf( "spu %u: all_reduce, wrong op %u\n", my_id, op ); 
    }

  /* reduce: we visit all four colums of the table */
  for( phase = 0; phase < 4; phase++ )
    {
      /* MAX_SPU, which is expected to be 16, identifies an idle
	 phase; a non-negative number means that the SPU sends to a
	 partner while a negative one means that we receive from a
	 partner */
      partner = binary_tree[my_id][phase];

      if ( partner >= 0 && partner < (int) spus )
	{
	  /* set the presence byte */
	  reduce_send_buffer.cval[15] = 1;

	  /* send the value */
	  blocking_put( (void*) &reduce_send_buffer, cb.ls[partner] + 
	  	((unsigned int) &reduce_recv_buffer[my_id] ), 
			sizeof( communication_buffer ), tag );
	}
      else if ( -((int) spus) < partner && partner < 0 )
	{
	  volatile communication_buffer* rbuf = 
	    (communication_buffer*) &reduce_recv_buffer[-partner];
	 
	  DB_RED(spe_printf( "spu %u, phase %u receiving from %u\n", 
			     my_id, phase, (unsigned int) -partner );)

	  /* wait for message arrival */
	  while ( rbuf->cval[15] == 0 );
	  
	  /* reset the buffer */
	  rbuf->cval[15] = 0;

	  DB_RED(spe_printf( "spu %u, phase %u received from %u\n", 
			     my_id, phase, (unsigned int) -partner );)

	  /* update the send buffer */
	  switch ( op )
	    {
	    case REDUCE_BARRIER:
	      break;
	    case REDUCE_INT:
	      reduce_send_buffer.ival[0] += rbuf->ival[0];
	      break;
	    case REDUCE_INT_MAX:
	      reduce_send_buffer.ival[0] = sw_max( reduce_send_buffer.ival[0],
						   rbuf->ival[0] );
	      break;
	    case REDUCE_DOUBLE:
	      reduce_send_buffer.dval[0] += rbuf->dval[0];
	      break;
	    case REDUCE_DOUBLE_MAX:
	      reduce_send_buffer.dval[0] = sw_max( reduce_send_buffer.dval[0],
						   rbuf->dval[0] );
	      break;
	    default:
	      spe_printf( "spu %u: all_reduce, wrong op %u\n", my_id, op ); 
	    }
	}
    }

  /* broadcast: the tree is visited in the reverse direction */
  for( phase = 3; phase >= 0; phase-- )
    {
      /* MAX_SPU identifies an idle phase; a negative number means
	 that the SPU sends to a partner while a positive one means
	 that we receive from a partner */
      partner = binary_tree[my_id][phase];

      if ( ((int) -spus) < partner && partner < 0 )
	{
	  /* set the presence byte */
	  reduce_send_buffer.cval[15] = 1;

	  /* send the value */
	  blocking_put( (void*) &reduce_send_buffer, cb.ls[-partner] + 
			((unsigned int) &reduce_recv_buffer[my_id]), 
			sizeof( communication_buffer ), tag );
	}
      else if ( partner >= 0 && partner < (int) spus )
	{
	  volatile communication_buffer* rbuf = 
	    (communication_buffer*) &reduce_recv_buffer[partner];

	  /* wait for message arrival */
	  while ( rbuf->cval[15] == 0 );

	  /* reset the buffer */
	  rbuf->cval[15] = 0;

	  /* update the send buffer and store the final result in the
	     input buffer */
	  switch ( op )
	    {
	    case REDUCE_BARRIER:
	      break;
	    case REDUCE_INT:
	    case REDUCE_INT_MAX:
	      reduce_send_buffer.ival[0] = rbuf->ival[0];
	      buffer->ival[0] = reduce_send_buffer.ival[0];
	      break;
	    case REDUCE_DOUBLE:
	    case REDUCE_DOUBLE_MAX:
	      reduce_send_buffer.dval[0] = rbuf->dval[0];
	      buffer->dval[0] = reduce_send_buffer.dval[0];
	      break;
	    default:
	      spe_printf( "spu %u: all_reduce, wrong op %u\n", my_id, op ); 
	    }
	}
    }

  /* the source set the return value */
  if ( !my_id )
    switch ( op )
      {
      case REDUCE_BARRIER:
	break;
      case REDUCE_INT:
      case REDUCE_INT_MAX:
	buffer->ival[0] = reduce_send_buffer.ival[0];
	break;
      case REDUCE_DOUBLE:
      case REDUCE_DOUBLE_MAX:
	buffer->dval[0] = reduce_send_buffer.dval[0];
	break;
      default:
	spe_printf( "spu %u: all_reduce, wrong op %u\n", my_id, op ); 
      }
  
  DB_RED(spe_printf("spu %u completed all_reduce\n", my_id );)
}


int all_reduce_int_light( int val, unsigned int tag ) 
{
  int phase, partner;
  communication_buffer buffer;

  /* set the initial value */
  reduce_send_buffer.ival[0] = val;

  /* set the presence byte */
  reduce_send_buffer.cval[15] = 1;

  for( phase = 0; phase < light_phases; phase++ )
    {
      partner = binomial_tree1[my_id][phase];
      
      volatile communication_buffer* rbuf = 
	(communication_buffer*) &reduce_recv_buffer[-partner];
      
      if ( spu_extract( reduce_cond1, phase ) )
	mfc_put( (void*) &reduce_send_buffer, cb.ls[partner] + 
		 ((unsigned int) &reduce_recv_buffer[my_id] ), 
		 sizeof( communication_buffer), tag, 0, 0); 
      else if ( spu_extract( reduce_cond2, phase ) )
	{
	  /* wait for message arrival */
	  while ( rbuf->cval[15] == 0 );
	  
	  /* reset the buffer */
	  rbuf->cval[15] = 0;
	  
	  reduce_send_buffer.ival[0] += rbuf->ival[0];
	}
    }
  
  /* broadcast: the tree is visited in the reverse direction */
  for( phase = light_phases -1; phase >= 0; phase-- )
    {
      partner = binomial_tree1[my_id][phase];
      
      volatile communication_buffer* rbuf = 
	(communication_buffer*) &reduce_recv_buffer[partner];

      if ( spu_extract( broadcast_cond1, phase ) )
	/* send the value */
	mfc_put( (void*) &reduce_send_buffer, cb.ls[-partner] + 
		 ((unsigned int) &reduce_recv_buffer[my_id] ), 
		 sizeof( communication_buffer ), tag, 0, 0);
      else if ( spu_extract( broadcast_cond2, phase ) )
	{
	  /* wait for message arrival */
	  while ( rbuf->cval[15] == 0 );
	  
	  /* reset the buffer */
	  rbuf->cval[15] = 0;
	  
	  reduce_send_buffer.ival[0] = rbuf->ival[0];
	  buffer.ival[0] = reduce_send_buffer.ival[0];
	}
    }
  
  /* the source set the return value */
  return ( !my_id ) ? reduce_send_buffer.ival[0]: buffer.ival[0];
}


void barrier( unsigned tag )
{
  all_reduce( 0, REDUCE_BARRIER, tag );
}


int all_reduce_int( int val, unsigned int tag )
{
  communication_buffer buffer;

  buffer.ival[0] = val;

  all_reduce( &buffer, REDUCE_INT, tag );

  return buffer.ival[0];
}


int all_reduce_int_max( int val, unsigned int tag )
{
  communication_buffer buffer;

  buffer.ival[0] = val;

  all_reduce( &buffer, REDUCE_INT_MAX, tag );

  return buffer.ival[0];
}


double all_reduce_double( double val, unsigned int tag )
{
  communication_buffer buffer;

  buffer.dval[0] = val;

  all_reduce( &buffer, REDUCE_DOUBLE, tag );

  return buffer.dval[0];
}


double all_reduce_double_max( double val, unsigned int tag )
{
  communication_buffer buffer;

  buffer.dval[0] = val;

  all_reduce( &buffer, REDUCE_DOUBLE_MAX, tag );

  return buffer.dval[0];
}

#else


void barrier( unsigned tag )
{
  unsigned i;

  if ( my_id == 0 )
    {
      /* SPE 0 waits for a signal from all other SPEs */
      for( i = 1; i < spus; i++ )
	{
	  volatile communication_buffer* rbuf = 
	    (communication_buffer*) &recv_buffer[i];

	  /* wait for message arrival */
	  while ( rbuf->ival[0] == 0 );
	  
	  /* reset the buffer */
	  rbuf->ival[0] = 0;
	}

      send_buffer.ival[0] = True;

      /* and sends back an ack */
      for( i = 1; i < spus; i++ )
	mfc_put( (void*) &send_buffer, cb.ls[i] + 
		 ((unsigned int) &recv_buffer[my_id] ), 
		 sizeof( communication_buffer), tag, 0, 0); 
      wait( tag );

      send_buffer.ival[0] = 0;
    }
  else
    {
      volatile communication_buffer* rbuf = 
	(communication_buffer*) &recv_buffer[0];

      send_buffer.ival[0] = True;

      /* send a message to SPE 0 */
      mfc_put( (void*) &send_buffer, cb.ls[0] + 
	       ((unsigned int) &recv_buffer[my_id] ), 
	       sizeof( communication_buffer), tag, 0, 0); 
      wait( tag );

      send_buffer.ival[0] = 0;

      /* and wait for the final ack */
      while ( rbuf->ival[0] == 0 );
	  
      /* reset the buffer */
      rbuf->ival[0] = 0;
    }
}


#endif /* COMMUNICATION_RUNTIME */


/*
*******************************************************************************
*
*                      initialization and monitoring
*
*******************************************************************************
*/


#ifdef DEBUG

void print_spe_control_block( unsigned long long speid )
{
  unsigned int i, cb_size = sizeof(spe_control_block);

  spe_printf( "spu %u: speid 0x%llx, spus %u, lg_spus %u "
	      "control block size %u\n", my_id, speid, spus, log_spus, 
	      cb_size );

  for( i = 0; i < spus; i++ )
    {
      spe_printf( "spu %d: spu %d, ls 0x%llx, mirror ls 0x%llx, "
		  "ps_cntl 0x%llx\n", 
		  my_id, i, cb.ls[i], cb.mirror_ls[i], cb.ps_cntl[i] );

      spe_printf( "spu %d: spu %d, smb 0x%llx inp 0x%llx, lgb 0x%llx "
		  "inp 0x%llx\n",
		  my_id, i, cb.small_buffer[i], cb.in_small_buffer[i], 
		  cb.large_buffer[i], cb.in_large_buffer[i] );
    }
}

#endif /* DEBUG */


void dump_large_buffer()
{
  uint64_t large_buffer = cb.large_buffer[my_id] + 
    remote_in_large * sizeof( unsigned int );
  uint64_t in_large_buffer = cb.in_large_buffer[my_id];
  unsigned int src_buffer = (unsigned int) spu_large_buffer;
  unsigned int chunk_size, msg_size = in_large * sizeof( unsigned int );

  /* round up the message size to match a 16-byte boundary: in all
     cases, but the last time, dump_large_buffer is called when the
     local buffer is full, and this is dy design aligned on a 16-byte
     boundary; for the last iteration, we round up to make the DMA
     controller happy */
  msg_size = ( msg_size + 0xF ) & -( 0x10 );

  if ( remote_in_large + in_large > LARGE_BUFFER_SIZE )
    spe_printf( "dump_large_buffer: large_buffer overflow %u vs %u\n",
		remote_in_large + in_large, LARGE_BUFFER_SIZE );

  while ( msg_size )
    {
      chunk_size = ( msg_size > MAX_DMA_SIZE ) ? MAX_DMA_SIZE: msg_size;

      /* dump the spu_large_buffer */
      blocking_put( (void*)src_buffer, large_buffer, chunk_size, 
		    DMA_TAG_BLOCKING );

      src_buffer += chunk_size;
      large_buffer += chunk_size;
      msg_size -= chunk_size;
    }

  /* update the insertion counters */
  remote_in_large += in_large;
  in_large = 0;

  blocking_put( &remote_in_large, in_large_buffer, sizeof(unsigned int), 
		DMA_TAG_BLOCKING );
}


void dump_small_buffer()
{
  uint64_t small_buffer = cb.small_buffer[my_id] + 
    remote_in_small * sizeof( unsigned int );
  uint64_t in_small_buffer = cb.in_small_buffer[my_id];

  if ( remote_in_small + in_small > SMALL_BUFFER_SIZE )
    spe_printf( "dump_small_buffer: small_buffer overflow %u vs %u\n",
		remote_in_small + in_small, SMALL_BUFFER_SIZE );

  /* update the remote insertion counter */
  remote_in_small += in_small;

  /* dump the spu_small_buffer */
  blocking_put( (void*) spu_small_buffer, small_buffer, 
		SPU_SMALL_BUFFER_SIZE * sizeof(unsigned int), 
		DMA_TAG_BLOCKING );

  /* reset the local counter */
  in_small = 0;

  blocking_put( (void*) &remote_in_small, in_small_buffer, 
		sizeof(unsigned int), DMA_TAG_BLOCKING );
}


inline void log_event( unsigned int val, unsigned int event_mask )
{
  /* when we have an overflow, we use three entries in the buffer
     (1) for the overflow event tag
     (2) for the event tag
     (3) the actual entry, that can use the full 32 bits */
  
  if (  __builtin_expect( ( val > EVENT_VALUE_BITMASK ), 0 ) )
    {
      if ( __builtin_expect( (in_large == SPU_LARGE_BUFFER_SIZE), 0 ) )
	dump_large_buffer();

      spu_large_buffer[in_large++] = OVERFLOW_EVENT;

      if ( __builtin_expect( (in_large == SPU_LARGE_BUFFER_SIZE), 0 ) )
	dump_large_buffer();
      spu_large_buffer[in_large++] = event_mask;

      if ( __builtin_expect( (in_large == SPU_LARGE_BUFFER_SIZE), 0 ) )
	dump_large_buffer();

      spu_large_buffer[in_large++] = val;
    }
  else
    {
      if ( __builtin_expect( (in_large == SPU_LARGE_BUFFER_SIZE), 0 ) )
	dump_large_buffer();
      
      spu_large_buffer[in_large++] = val | event_mask;
    }
}


void initialize_spu_thread( addr64 argp )
{
  /* wait for the start message from the PPU, that contains the SPU id */
  my_id = spu_readch( SPU_RdInMbox );

  /* get the control block */
  blocking_get( &cb, argp.ull, CB_SIZE, DMA_TAG_BLOCKING );
 
 /* set the number of SPUs */
  spus = cb.spus;
  log_spus = cb.log_spus;

  /* print the control block */
  DB(print_spe_control_block( speid );)

#ifdef COMMUNICATION_RUNTIME
  /* initialize the communication channels */  
  init_point_to_point();
  init_collectives();
#else
  /* clean the send and receive buffers */
  spe_memset( (void*) &send_buffer, 0, sizeof( communication_buffer ) );
  spe_memset( (void*) &recv_buffer[0], 0, 
	      MAX_SPU * sizeof( communication_buffer ) );

#endif /* COMMUNICATION_RUNTIME */

  /* send an ack back to the PPU with the SPU id */
  spu_writech( SPU_WrOutMbox, my_id );

  /* and wait for the final message from the PPU */
  my_id = spu_readch( SPU_RdInMbox);
}


