/*
*******************************************************************************
*
* File:         runtime_spu.h
* RCS:          $Header: $
* Description:  The SPU communication runtime support
* Author:       Fabrizio Petrini 
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/


#ifndef __runtime_spu__
#define __runtime_spu__

#ifdef LIBSPE1

#include <profile.h>

#define PROF_CLEAR  prof_clear()
#define PROF_START  prof_start()
#define PROF_STOP   prof_stop()

#else /* LIBSPE2 */

#define PROF_CLEAR
#define PROF_START
#define PROF_STOP

#endif /* LIBSPE1 */

#include <stdio.h>
#include <spu_mfcio.h>
#include <libsync.h>
#include "../common.h"


/*
*******************************************************************************
*
*                               defines
*
*******************************************************************************
*/


/* the initialization value of the decrementer register */
#define DECR_COUNT 0xFFFFFFFF

/* DMA tag for blocking communication */
#define DMA_TAG_BLOCKING 30

/* alignment constant */
#define DMA_ALIGN 7


/*
*******************************************************************************
*
*                              initialization
*
*******************************************************************************
*/


/* my SPU id and the number of spus; initialized by initialize_spu_thread */
extern unsigned int my_id;
extern unsigned int spus;
extern unsigned int log_spus;

extern volatile communication_buffer recv_buffer[MAX_SPU] align128;
extern communication_buffer send_buffer align128;


/* we allocate one control block, that contains the info about the threads */
extern spe_control_block cb;

/* initialization function: it is called by the SPU thread at the very
   beginning; takes the input arguments */
void initialize_spu_thread( addr64 argp );


/*
*******************************************************************************
*
*                                monitoring 
*
*******************************************************************************
*/


/* local and remote in pointers for the small and large buffers */
extern unsigned int in_small;
extern unsigned int in_large; 
extern unsigned int remote_in_small;
extern unsigned int remote_in_large;

/* small and large buffers */
extern unsigned int spu_small_buffer[SPU_SMALL_BUFFER_SIZE];
extern unsigned int spu_large_buffer[SPU_LARGE_BUFFER_SIZE];

/* log an event: take the event value and the event identifier */
void log_event( unsigned int val, unsigned int event_mask );

/* lightweight version of log_event: if we already know that there is
   no spu_large_buffer overflow, we can use this function; to be used
   with extreme care */
#define light_log_event( val, event_mask )		\
  spu_large_buffer[in_large++] = val | event_mask;

/* dump the monitoring buffer */
void dump_large_buffer();


/*
*******************************************************************************
*
*                       communication and synchronization 
*
*******************************************************************************
*/


#ifdef COMMUNICATION_RUNTIME

/* dmas */
void print_put( unsigned ls, unsigned ea, unsigned size, unsigned tag );
void print_get( unsigned ls, unsigned ea, unsigned size, unsigned tag );

void blocking_get( void* ls, uint64_t rs, unsigned size, unsigned tag );
void large_blocking_get( void* ls, uint64_t rs, unsigned size, unsigned tag );
void blocking_put( void* ls, uint64_t rs, unsigned size, unsigned tag );
void large_blocking_put( void* ls, uint64_t rs, unsigned size, unsigned tag );

/* wait functions */
void wait (unsigned int tag);
unsigned wait_immediate( unsigned wait_mask );
unsigned wait_any( unsigned wait_mask );
boolean_t tag_is_outstanding( unsigned dma_tag );

/* blocking communication primitives */
void blocking_send( unsigned int receiver, void* send_buffer, 
		    unsigned int size, unsigned int tag );

void blocking_recv( unsigned int sender, void* recv_buffer, 
		    unsigned int size, unsigned int tag );

/* collective communication: various flavors of single-element all-reduces */
void barrier(unsigned tag);
int all_reduce_int( int val, unsigned int tag );
int all_reduce_int_max( int val, unsigned int tag );
double all_reduce_double( double val, unsigned int tag );
double all_reduce_double_max( double val, unsigned int tag );

/* these functions must be called until they return TRUE; they keep
   forwarding messages in the reduction and distribution trees,
   respectively, until they complete their task */
int sync_reduce( unsigned int tag );
int sync_broadcast( unsigned int tag );

/* verify the correctness of the dma */
void check_dma_align( void* ls, uint64_t rs, unsigned int size );


#else


/* minimal communication interface */

#define wait( tag )				\
  mfc_write_tag_mask( 1 << tag );		\
  mfc_read_tag_status_all();

/* dma put */
#define blocking_put( ls, rs, size, tag )		\
  mfc_put( (void*) ls, (uint64_t) rs, size, tag, 0, 0);	\
  mfc_write_tag_mask( 1 << tag );			\
  mfc_read_tag_status_all();

/* dma get */
#define blocking_get( ls, rs, size, tag )		\
  mfc_get( (void*) ls, (uint64_t) rs, size, tag, 0, 0);	\
  mfc_write_tag_mask( 1 << tag );			\
  mfc_read_tag_status_all();

/* barrier synchronization */
void barrier( unsigned tag );


#endif /* COMMUNICATION_RUNTIME */


#endif /* __runtime_spu__ */
